//
//  AQPlayFileModel2.m
//  AudioQueuePlayFile
//
//  Copyright 2011 JG Electronics. All rights reserved.
//	Remarks, reports to: www.jgelectronics.nl.
//
//	Derived from Apple's example program AudioQueueTools.
//	The application icon is from Vincent Garnier http://benjigarner.deviantart.com/art/Fruity-Apples-80805804
//
//	OSX10.6 and XCODE 4.0.1
//
/*
 Disclaimer: IMPORTANT: 
 This JG Electronics software is supplied to you by JG Electronics as FREE AND OPEN SOURCE SOFTWARE. 
 
 The JG Electronics Software is provided by JG Electronics on an "AS IS" basis.
 JG ELECTRONICS MAKES NO WARRANTIES, EXPRESS OR IMPLIED, INCLUDING WITHOUT LIMITATION
 THE IMPLIED WARRANTIES OF NON-INFRINGEMENT, MERCHANTABILITY AND FITNESS
 FOR A PARTICULAR PURPOSE, REGARDING THE JG ELECTRONICS SOFTWARE OR ITS USE AND
 OPERATION ALONE OR IN COMBINATION WITH YOUR PRODUCTS.
 
 IN NO EVENT SHALL JG ELECTRONICS BE LIABLE FOR ANY SPECIAL, INDIRECT, INCIDENTAL
 OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) ARISING IN ANY WAY OUT OF THE USE, REPRODUCTION,
 MODIFICATION AND/OR DISTRIBUTION OF THE JG ELECTRONICS SOFTWARE, HOWEVER CAUSED
 AND WHETHER UNDER THEORY OF CONTRACT, TORT (INCLUDING NEGLIGENCE),
 STRICT LIABILITY OR OTHERWISE, EVEN IF JG ELECTRONICS HAS BEEN ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
 */
#import "AQPlayFileModel2.h"

static UInt32		gIsRunning = 0;

@implementation AQPlayFileModel2

	#pragma mark -
	#pragma mark === Internal Methods ===
//	*********************************************************************************************************************************
- (id)init
{
    self = [super init];
    if (self) {
		
    }
    
    return self;
}

- (void)dealloc
{
    [super dealloc];
}

	#pragma mark -
	#pragma mark === Callback Procedures ===
//	*********************************************************************************************************************************
static void AQTestBufferCallback(void *	aqData, AudioQueueRef inQueue, AudioQueueBufferRef	inBuffer) 
{
	AQPlayFileModel2 * pAqData = (AQPlayFileModel2 *)aqData;
//	AQTestInfo * myInfo = (AQTestInfo *)inUserData;
	
	if (pAqData->mDone) return;
	
	UInt32		numBytes;
	UInt32		nPackets = pAqData->mNumPacketsToRead;
	OSStatus	error = noErr;
	
	error = AudioFileReadPackets(pAqData->mAudioFile, false, &numBytes, pAqData->mPacketDescs, pAqData->mCurrentPacket, &nPackets, inBuffer->mAudioData);
	if(error != noErr) {
		NSLog(@"Reading from file FAILED!");
	//	DebugMessageN1 ("Error reading from file: %d\n", (int)result);
	//	exit(1);
	}
	
	if (nPackets > 0) {
		inBuffer->mAudioDataByteSize = numBytes;		
		
		AudioQueueEnqueueBuffer(inQueue, inBuffer, (pAqData->mPacketDescs ? nPackets : 0), pAqData->mPacketDescs);
		
		pAqData->mCurrentPacket += nPackets;
	} else {
		error = AudioQueueStop(pAqData->mQueue, false);
		if(error != noErr) {
			NSLog(@"AudioQueueStop FAILED!");
		}
	//	reading nPackets == 0 is our EOF condition
		pAqData->mDone = true;
	}
}

void MyAudioQueuePropertyListenerProc(void * inUserData, AudioQueueRef inQueue, AudioQueuePropertyID inID)
{	
	NSLog(@"MyAudioQueuePropertyListenerProc Start");
	UInt32		size = sizeof(gIsRunning);
	OSStatus	error = AudioQueueGetProperty(inQueue, kAudioQueueProperty_IsRunning, &gIsRunning, &size);
	if(error != noErr) {
		gIsRunning = 0;
	}
	if(gIsRunning != true) {
	//	local autorelease pool because this is a secondary thread and we call a cocoa method from this thread
	//	NSAutoreleasePool *pool = [[NSAutoreleasePool alloc] init];
		NSAutoreleasePool *pool = [NSAutoreleasePool new];
		[[NSNotificationCenter defaultCenter] postNotificationName:@"PlayBackStop" object:nil];
		[pool release];
	}
}

	#pragma mark -
	#pragma mark === Audio Methods ===
//	*********************************************************************************************************************************
//	we only use time here as a guideline
//	we're really trying to get somewhere between 16K and 64K buffers, but not allocate too much if we don't need it
- (UInt32)pfCalculateBytesForTime:(AudioStreamBasicDescription)inDesc maxPacketSize:(UInt32)inMaxPacketSize seconds:(Float64)inSeconds
{
	NSLog(@"CalculateBytesForTime Start");
	UInt32		outBufferSize;
	static const int maxBufferSize = 0x10000;	//	limit size to 64K
	static const int minBufferSize = 0x4000;	//	limit size to 16K
	
	if (inDesc.mFramesPerPacket) {
		Float64 numPacketsForTime = inDesc.mSampleRate / inDesc.mFramesPerPacket * inSeconds;
		outBufferSize = numPacketsForTime * inMaxPacketSize;
	} else {
	//	if frames per packet is zero, then the codec has no predictable packet == time
	//	so we can't tailor this (we don't know how many Packets represent a time period
	//	we'll just return a default buffer size
		outBufferSize = maxBufferSize > inMaxPacketSize ? maxBufferSize : inMaxPacketSize;
	}
//	we're going to limit our size to our default
	if (outBufferSize > maxBufferSize && outBufferSize > inMaxPacketSize)
		outBufferSize = maxBufferSize;
	else {
	//	also make sure we're not too small - we don't want to go the disk for too small chunks
		if (outBufferSize < minBufferSize)
			outBufferSize = minBufferSize;
	}
	mNumPacketsToRead = outBufferSize / inMaxPacketSize;
	return outBufferSize;
}

- (BOOL)pfConfigureAudioFile2:(CFURLRef)inURL
{
	Float32			volume = 0.75;
	OSStatus		error = noErr;
	
	error = AudioFileOpenURL (inURL, kAudioFileReadPermission , 0, &mAudioFile);
	if(error != noErr) {
		NSLog(@"AudioFileOpenURL failed");
	}
		
	UInt32	size;
	error =	AudioFileGetPropertyInfo(mAudioFile, kAudioFilePropertyFormatList, &size, NULL);
	if(error != noErr) {
		NSLog(@"AudioFileGetPropertyInfo size failed");
	}
	UInt32	numFormats = size / sizeof(AudioFormatListItem);
	
//	AudioFormatListItem *	formatList = (AudioFormatListItem *) new AudioFormatListItem[numFormats];	//	this is a C++ expression
//	formatList is a AudioFormatListItem pointer to an array of AudioFormatListItems containing numFormats items
//	To let the compiler accept C++ we must use file extension .mm
//	If we want to use file extension .m we have to change the expression to a C expression
	AudioFormatListItem *	formatList = (AudioFormatListItem *) malloc(numFormats * sizeof(AudioFormatListItem));
//	If we no longer need the formatlist we must free the memory
	
	error =	AudioFileGetProperty(mAudioFile, kAudioFilePropertyFormatList, &size, formatList);
	if(error != noErr) {
		NSLog(@"AudioFileGetProperty formatList failed");
	}
	numFormats = size / sizeof(AudioFormatListItem); // we need to reassess the actual number of formats when we get it
		
	if (numFormats == 1) {
//	this is the common case
		mDataFormat = formatList[0].mASBD;
	//	see if there is a channel layout (multichannel file)
		error = AudioFileGetPropertyInfo(mAudioFile, kAudioFilePropertyChannelLayout, &mChannelLayoutSize, NULL);
			if (error == noErr && mChannelLayoutSize > 0) {
			//	mChannelLayout = (AudioChannelLayout *)new char [mChannelLayoutSize];	//	C++ expression change to C
				mChannelLayout = (AudioChannelLayout *) malloc(mChannelLayoutSize * sizeof(char));
				error =	AudioFileGetProperty(mAudioFile, kAudioFilePropertyChannelLayout, &mChannelLayoutSize, mChannelLayout);
				if(error != noErr) {
					NSLog(@"AudioFileGetProperty mChannelLayout failed");
				}
			}
		} else {
		//	now we should look to see which decoders we have on the system
			error = AudioFormatGetPropertyInfo(kAudioFormatProperty_DecodeFormatIDs, 0, NULL, &size);
			if(error != noErr) {
				NSLog(@"AudioFormatGetPropertyInfo decoderIDs size failed");
			}
			UInt32		numDecoders = size / sizeof(OSType);
//			OSType *	decoderIDs = new OSType [numDecoders];	//	C++ expression change to C
			OSType *	decoderIDs = (OSType*) malloc(numDecoders  * sizeof(OSType));
			error = AudioFormatGetProperty(kAudioFormatProperty_DecodeFormatIDs, 0, NULL, &size, decoderIDs);
			if(error != noErr) {
				NSLog(@"AudioFormatGetProperty decoderIDs failed");
			}
			unsigned int i = 0;
			for (; i < numFormats; ++i) {
				OSType decoderID = formatList[i].mASBD.mFormatID;
				bool found = false;
				for (unsigned int j = 0; j < numDecoders; ++j) {
					if (decoderID == decoderIDs[j]) {
						found = true;
						break;
					}
				}
				if (found) break;
			}
//			delete [] decoderIDs;	//	C++ expression change to C
			free(decoderIDs);
			
			if (i >= numFormats) {
				NSLog(@"Cannot play any of the formats in this file");
//				throw kAudioFileUnsupportedDataFormatError;
			}
			mDataFormat = formatList[i].mASBD;
			mChannelLayoutSize = sizeof(AudioChannelLayout);
		//	mChannelLayout = (AudioChannelLayout*)new char [mChannelLayoutSize];	//	C++ expression change to C
			mChannelLayout = (AudioChannelLayout*) malloc(mChannelLayoutSize * sizeof(char));
			mChannelLayout->mChannelLayoutTag = formatList[i].mChannelLayoutTag;
			mChannelLayout->mChannelBitmap = 0;
			mChannelLayout->mNumberChannelDescriptions = 0;
		}
//		delete [] formatList;	//	C++ expression change to C
		free(formatList);		//	free formatlist memory
	
		error =	AudioQueueNewOutput(&mDataFormat, AQTestBufferCallback, self, CFRunLoopGetCurrent(), kCFRunLoopCommonModes, 0, &mQueue);
		if(error != noErr) {
			NSLog(@"AudioQueueNew failed");
		}
	//	we need to calculate how many packets we read at a time, and how big a buffer we need
	//	we base this on the size of the packets in the file and an approximate duration for each buffer
		bool isFormatVBR = (mDataFormat.mBytesPerPacket == 0 || mDataFormat.mFramesPerPacket == 0);
	//	first check to see what the max size of a packet is - if it is bigger
	//	than our allocation default size, that needs to become larger
		UInt32 maxPacketSize;
		size = sizeof(maxPacketSize);
		error =	AudioFileGetProperty(mAudioFile, kAudioFilePropertyPacketSizeUpperBound, &size, &maxPacketSize);
		if(error != noErr) {
			NSLog(@"AudioFileGetProperty maxPacketSize failed");
		}
	//	adjust buffer size to represent about a half second of audio based on this format
//		CalculateBytesForTime (mDataFormat, maxPacketSize, 0.5/*seconds*/, &bufferByteSize, &mNumPacketsToRead);
		UInt32 bufferByteSize = [self pfCalculateBytesForTime:mDataFormat maxPacketSize:maxPacketSize seconds:0.5];
		
		if (isFormatVBR)
		//	mPacketDescs = new AudioStreamPacketDescription [mNumPacketsToRead];
			mPacketDescs = (AudioStreamPacketDescription *) malloc(mNumPacketsToRead * sizeof(AudioStreamPacketDescription));
		else
			mPacketDescs = NULL;	//	we don't provide packet descriptions for constant bit rate formats (like linear PCM)

	//	If the file has a cookie, we should get it and set it on the Audio Queue
		size = sizeof(UInt32);
		error = AudioFileGetPropertyInfo (mAudioFile, kAudioFilePropertyMagicCookieData, &size, NULL);
		
		if (!error && size) {
		//	char* cookie = new char [size];
			char *	cookie = (char *) malloc(size);
			error =	AudioFileGetProperty (mAudioFile, kAudioFilePropertyMagicCookieData, &size, cookie);
			if(error != noErr) {
				NSLog(@"AudioFileGetProperty get cookie failed");
			}
			error =	AudioQueueSetProperty(mQueue, kAudioQueueProperty_MagicCookie, cookie, size);
			if(error != noErr) {
				NSLog(@"AudioQueueSetProperty set cookie failed");
			}
			free (cookie);
		}
		
//	set Audio Channel Layout if there is one
	if (mChannelLayout) {
		error =	AudioQueueSetProperty(mQueue, kAudioQueueProperty_ChannelLayout, mChannelLayout, mChannelLayoutSize);
		if(error != noErr) {
			NSLog(@"AudioQueueSetProperty mChannelLayoutSize failed");
		}
	}
//	prime the queue with some data before starting
	mDone = false;
	mCurrentPacket = 0;
	for (int i = 0; i < kNumberBuffers; ++i) {
		error =	AudioQueueAllocateBuffer(mQueue, bufferByteSize, &mBuffers[i]);
		if(error != noErr) {
			NSLog(@"AudioQueueAllocateBuffer failed");
		}
		AQTestBufferCallback (self, mQueue, mBuffers[i]);
		if (mDone) break;
	}	
//	set the volume of the queue
	error =	AudioQueueSetParameter(mQueue, kAudioQueueParam_Volume, volume);
	if(error != noErr) {
		NSLog(@"AudioQueueSetParameter volume failed");
	}
	error =	AudioQueueAddPropertyListener (mQueue, kAudioQueueProperty_IsRunning, MyAudioQueuePropertyListenerProc, NULL);
	if(error != noErr) {
		NSLog(@"AudioQueueAddPropertyListener failed");
	}
	return YES;
}

	#pragma mark -
	#pragma mark === Audio Control Methods ===
//	*********************************************************************************************************************************
- (void)fpSetamplitude:(float)amplFlt {
	OSStatus	error = noErr;
	
		//	set the volume of the queue
	error =	AudioQueueSetParameter(mQueue, kAudioQueueParam_Volume, amplFlt);
	if(error != noErr) {
		NSLog(@"AudioQueueSetParameter volume failed");
	}
}

- (BOOL)pfStartPlayBack
{
	OSStatus	error = noErr;
	
	error =	AudioQueueStart(mQueue, NULL); 
	if(error != noErr) {
		NSLog(@"AudioQueueStart failed");
		return NO;
	}
	return YES;
}

- (BOOL)pfStopPlayBack
{
	NSLog(@"StopPlayBack Start");
	OSStatus	error = noErr;
	
	error =	AudioQueueDispose(mQueue, true);
	if(error != noErr) {
		NSLog(@"AudioQueueDispose failed");
		return NO;
	}

	error =	AudioFileClose(mAudioFile);
	if(error != noErr) {
		NSLog(@"AudioQueueDispose(false) failed");
		return NO;
	}
	return YES;
}

@end
